! exprparse  -  parse arithmetic and logical expressions

      logical function exprparse (t, token, av, exprtype)
      integer        t, av, exprtype
      character*(*)  token
      #include       global.def
      #include       spl.mac
      #include       vm.mac
      #include       vstate.cmn
      integer        class, c, ptok, lex, symalloc, i
      logical        symref
      integer        attr (NATTR)
      #include       parse.cmn
! parse tables
      integer  iprecedence (MAXTOKEN), sprecedence (MAXTOKEN)
      logical  allowed (MAXCLASS, MAXCLASS)
      #include exprparse.def

! initialize and get initial token
      esp = 0
      psp = 0
      ptok = OTHER
      osp = 1
      opclass(1) = OTHER
      optoken(1) = $eol
      c = class(ptok, t, token, attr)
      if (c .eq. OTHER) then
         t = lex(t, token)
      end if
      c = class (ptok, t, token, attr)
      if (c .eq. OTHER) then
         call perror ('invalid expression'//NULL)
         do while (t .ne. $eol)
            t = lex(t,token)
         end do
         exprparse = .false.
         return
      end if

! read and parse expression
      #define endcondition &
      c .eq. OTHER .or. (c .eq. COMMA_SEPARATOR .and. psp .eq. 0)

      do while (.not.(endcondition))
         !check for errors
         if (.not. allowed(c, ptok)) then
            call perror ('invalid expression syntax'//NULL)
            do while (t .ne. $eol)
               t = lex(t,token)
            end do
            return
         else if (esp .eq. MAXSTACK .or. 
     &            osp .eq. MAXSTACK .or. 
     &            psp .eq. MAXSTACK) then
            call perror ('expression stack overflow'//NULL)
            do while (t .ne. $eol)
               t = lex(t,token)
            end do
            return
         end if

         ! process current token
         if (c .eq. CONSTANT .or. c .eq. IDENTIFIER .or. 
     &       c .eq. ARRAY_ID .or. c .eq. STRING_ID .or. 
     &       c .eq. FUNCTION_ID) then
            call operand (c, t, token, attr)
            if (c .eq. ARRAY_ID .or. c .eq. FUNCTION_ID) then
               osp = osp + 1
               optoken(osp) = t
               opclass(osp) = c
               org(osp) = attr(IDORG)
            end if
         else if (c .eq. OPEN_PAREN) then
            psp = psp + 1
            depth(psp) = esp
            prevdepth(psp) = esp
            osp = osp + 1
            opclass(osp) = OPEN_PAREN
            optoken(osp) = t
         else if (c .eq. COMMA_SEPARATOR) then
            do while (opclass(osp) .ne. OPEN_PAREN .and. osp .gt. 1)
               call operator (optoken(osp), opclass(osp))
               osp = osp - 1
            end do
            if (opclass(osp) .ne. OPEN_PAREN) then
               call perror ('missing parentheses in argument list'//NULL)
            else if (opclass(osp-1) .ne. FUNCTION_ID) then
               call perror ('allowed only in function argument list'//NULL)
            else if (esp - prevdepth(psp) .ne. 1) then
               call perror ('invalid argument expression'//NULL)
            else
               prevdepth(psp) = esp
            end if
            if (aorv(esp) .eq. VALUE) then  !allocate temporary address
               call emit2 (_push, symalloc(), INTTYPE)
               call emit1 (_asna)
            end if
         else if (c .eq. CLOSE_PAREN) then
            do while (opclass(osp) .ne. OPEN_PAREN .and. osp .gt. 1)
               call operator (optoken(osp),opclass(osp))
               osp = osp - 1
            end do
            if (opclass(osp) .eq. OPEN_PAREN) then
               osp = osp - 1          !discard opening parentheses
               if (opclass(osp) .eq. ARRAY_ID) then
                  if (esp - depth(psp) .ne. 1) then
                     call perror ('invalid subscript expression'//NULL)
                  else
                     if (type(esp) .eq. INTTYPE) then
                        if (aorv(esp) .eq. ADDRESS) call emit1 (_val)
                        call emit1 (_ixa)
                        esp = esp - 1
                        aorv(esp) = ADDRESS
                     else
                        call perror ('array index must be integer'//NULL)
                     end if
                  end if
                  osp = osp - 1       !discard ARRAY_ID
               else if (opclass(osp) .eq. FUNCTION_ID) then
                  if (esp .ne. depth(psp)) then  !arguments on stack
                     if (esp - prevdepth(psp) .ne. 1) then
                        call perror ('invalid argument expression'//NULL)
                     end if
                     if (aorv(esp) .eq. VALUE) then
                        call emit2 (_push, symalloc(), INTTYPE)
                        call emit1 (_asna)
                     end if
                     esp = depth(psp)
                  end if
                  call emit2 (_call, org(osp), INTTYPE)
               else           ! (expression)
                  if (esp - depth(psp) .ne. 1) then
                     call perror ('invalid expression'//NULL)
                     if (esp - depth(psp) .eq. 0) then
                        esp = esp + 1    !fake it for operator
                        type(esp) = INTTYPE
                        aorv(esp) = VALUE
                     end if
                  end if
               end if
               psp = psp - 1
            else
               call perror ('missing left parentheses'//NULL)
            end if
         else if (iprecedence(t) .gt. sprecedence(optoken(osp))) then
            if (osp .eq. MAXSTACK) then
               call perror ('operator stack overflow'//EOL)
               do while (t .ne. $eol)
                  t = lex(t,token)
               end do
            else
               osp = osp + 1
               opclass(osp) = c
               optoken(osp) = t
            end if
         else
            do while (iprecedence(t) .le. sprecedence(optoken(osp)))
               call operator (optoken(osp),opclass(osp))
               osp = osp - 1
            end do
            osp = osp + 1
            opclass(osp) = c
            optoken(osp) = t
         end if
         ptok = c
         t = lex(t,token)
         c = class(ptok,t,token,attr)
      end do

      do while (osp .gt. 1)        !unstack any remaining operators
         if (opclass(osp) .eq. OPEN_PAREN) then
            call perror ('missing right parentheses'//NULL)
         else
            call operator (optoken(osp), opclass(osp))
         end if
         osp = osp - 1
      end do

      exprparse = (esp .eq. 1)
      exprtype = type(1)
      av = aorv(1)

      return
      end
