! lexgen  -  lexical analyzer table generator

      program lexgen
      include            'global.def'
      include            'nfa.cmn'
      include            'dfa.cmn'
      integer            argc, uin, uout
      character*(MAXSTR) argv(MAXARGS), line, itoa, upper
      integer            startstate(256)
      integer            i, j, lp, ival, s, olds, nstates, nentries
      logical            ppline, is_ival, caseflag, errflag

      call ioinit ()
      call getargs (argc, argv)
      call stdargs (argc, argv, uin, uout)
      if (uin .eq. IOERROR .or. uout .eq. IOERROR) stop ' '
      caseflag = .false.
      do i = 1, argc
         if (argv(i)(1:1) .eq. QUALIFIER) then
            argv(i) = upper(argv(i))
            j = index(argv(i),NULL)
            if (index(argv(i)(1:j), 'U') .gt. 0) caseflag = .true.
         end if
      end do
      call ppinit (uin)

! initialize NFA
      do i = 1, MAXNFA
         chrset(i) = NONE
         cval1(i) = 0
         cval2(i) = 0
         nfatoken(i) = 0
      end do

! process all input  -  reserve initial state of NFA
      call fputstr (STDOUT, 'constructing NFA...'//EOL)
      i = 1
      s = 2
      epstabp = 1
      errflag = .false.
      do while (ppline (line))
         lp = 1
         do while (line(lp:lp) .eq. BLANK)
            lp = lp + 1
         end do
         if (line(lp:lp) .eq. NEWLINE .or. line(lp:lp) .eq. COMMENT) then
            continue
         else if (is_ival (ival, line, lp)) then
            do while (line(lp:lp) .eq. BLANK)
               lp = lp + 1
            end do
            if (line(lp:lp) .ne. '=') then
               call fputstr (STDERR, 'missing assignment operator'//EOL)
               call fputstr (STDERR, line)
            else
               lp = lp + 1
               startstate(i) = s
               i = i + 1
               olds = s
               call mknfa (s, line, lp)
               if (s .eq. olds) then
                  call fputstr (STDERR, 
     &            'no NFA states generated for token '//itoa(ival))
                  call fputc (STDERR, NEWLINE)
                  errflag = .true.
               else
                  nfatoken(s) = ival
               end if
               s = s + 1
            end if
         else
            call fputstr (STDERR, 'invalid token value'//EOL)
            call fputstr (STDERR, line)
         end if
      end do
      s = s - 1
      call fputstr (STDOUT, '# NFA states = '//itoa(s))
      call fputc (STDOUT, NEWLINE)
      if (errflag) stop 'lexgen aborted due to NFA errors.'

! link startstates together with epsset(1)
      startstate(i) = ENDLIST
      epsset(1) = epstabp
      call ascopy (startstate, epstabp, epstab)

! convert NFA to upper case if casefeflag set
      if (caseflag) then
         call fputstr (STDOUT, 'converting NFA to upper case...'//EOL)
         do i = 1, s
            if (cval1(i) .ge. ichar('a') .and. 
     &          cval1(i) .le. ichar('z')) then
               cval1(i) = cval1(i) - ichar('a') + ichar('A')
            end if
            if (cval2(i) .ge. ichar('a') .and. 
     &          cval2(i) .le. ichar('z')) then
               cval2(i) = cval2(i) - ichar('a') + ichar('A')
            end if
         end do
      end if

! convert NFA to DFA
      call fputstr (STDOUT, 'converting NFA to DFA...'//EOL)
      call mkdfa ()
      call fputstr (STDOUT, 'number of DFA states = '//itoa(d))
      call fputc (STDOUT, NEWLINE)
      if (dfatoken(1) .ne. 0) then
         call fputstr (STDERR, 'token value '//itoa(dfatoken(1)))
         call fputstr (STDERR, ' requires no input'//EOL)
         stop 'lexgen aborted due to DFA error.'
      end if

! optimize DFA
      call fputstr (STDOUT, 'optimizing DFA...'//EOL)
      call mkrfa ()
      call fputstr (STDOUT, 'number of DFA states = '//itoa(d))
      call fputc (STDOUT, NEWLINE)

! choose default states
      call fputstr (STDOUT, 'choosing default states...'//EOL)
      call mkdflt ()

! construct compacted table
      call fputstr (STDOUT, 'creating compacted table...'//EOL)
      call mklex (nstates, nentries)

! write output
      call fputstr (STDOUT, 'writing output...'//EOL)
      call lexdump (uout, nstates, nentries)

      end
