! mkrfa  -  minimize DFA state transition table (Mealy-Huffman algorithm)

      subroutine mkrfa ()
      include 'dfa.cmn'
      integer G (0:MAXDFA)             !current group of each DFA state
      integer Gtoken (MAXDFA)          !token value for group
      integer oldG (0:MAXDFA)          !previous group of each state
      integer nextG(0:MAXCHAR, MAXDFA) !next group for each input character
      integer group (MAXDFA)           !list of states in current group
      integer r                        !next available group number
      integer i, j, k, c, n
      logical eqvarray, same, found

! initialize groups
      r = 1                               !r=1 for nonaccepting states
      G(0) = 0                            !bogus dead state
      do i = 1, d
         G(i) = 0
         oldG(i) = 0
      end do
      do i = 1, d
         if (G(i) .ne. 0) then            !group already set
            continue
         else
            if (dfatoken(i) .eq. 0) then  !nonaccepting state
               G(i) = 1
            else                          !accepting state
               r = r + 1
               G(i) = r
               do j = i, d                !find other states with same token
                  if (dfatoken(i) .eq. dfatoken(j)) then
                     G(j) = G(i)
                  end if
               end do
            end if
         end if
      end do

! subdivide groups until termination
! terminate when two consecutive partitions are identical

      do while (.not. eqvarray (G, oldG, d))
         !record target groups of all states and update oldG
         do i = 1, d
            oldG(i) = G(i)
            do j = 0, MAXCHAR
               nextG(j,i) = G(dfanext(j,i))
            end do
         end do

         !check states in each group for equivalence
         do k = 1, r
            n = 0
            do j = 1, d           !make list of states in group k
               if (G(j) .eq. k) then
                  n = n + 1
                  group(n) = j
               end if
            end do
            do i = 2, n
               j = 0
               found = .false.
               do while ((j .lt. i-1) .and. (.not. found))
                  j = j + 1
                  same = .true.
                  c = 0
                  do while (c .le. MAXCHAR .and. same)
                     if (nextG(c,group(j)) .ne. nextG(c,group(i))) then
                        same = .false.
                     end if
                     c = c + 1
                  end do
                  if (same) found = .true.
               end do
               if (.not. found) then      !create new group
                  r = r + 1
                  G(group(i)) = r
               else                       !make same as matched group
                  G(group(i)) = G(group(j))
               end if
            end do
         end do                           !for all groups
      end do                              !until partition does not change

! reduced table is contained in nextG - copy to common block
      do i = 1, r
         do j = 1, d
            if (G(j) .eq. i) then        !state j is in group i
               Gtoken(i) = dfatoken(j)   !adopt the token value of state j
               do k = 0, MAXCHAR         !use group transitions of state j
                  dfanext(k,i) = nextG(k,j)
               end do
            end if
         end do
      end do

! set token values for reduced table
      do i = 1, r
         dfatoken(i) = Gtoken(i)
      end do
      d = r

      return
      end
